@@ -123,22 +123,17 @@ nm_l3_acd_addr_info_find_track_info(const NML3AcdAddrInfo *addr_info,
 }
 
 typedef enum {
-    /* emitted when the merged/commited NML3ConfigData instance changes.
+    NM_L3_CONFIG_NOTIFY_TYPE_ACD_EVENT,
+
+    /* Emitted before the merged l3cd is committed to platform.
      * Note that this gets emitted "under unsafe circumstances". That means,
      * you should not perform complex operations inside this callback,
      * and neither should you call into NML3Cfg again (reentrancy). */
-    NM_L3_CONFIG_NOTIFY_TYPE_L3CD_CHANGED,
-
-    NM_L3_CONFIG_NOTIFY_TYPE_ACD_EVENT,
-
-    /* emitted before the merged l3cd is committed to platform.
-     *
-     * This event also gets emitted "under unsafe circumstances".
-     * See NM_L3_CONFIG_NOTIFY_TYPE_L3CD_CHANGED. */
     NM_L3_CONFIG_NOTIFY_TYPE_PRE_COMMIT,
 
     /* emitted at the end of nm_l3cfg_platform_commit(). This signals also that
-     * nm_l3cfg_is_ready() might have switched to TRUE. */
+     * nm_l3cfg_is_ready() might have switched to TRUE. Also emitted
+     * "under unsafe circumstances". */
     NM_L3_CONFIG_NOTIFY_TYPE_POST_COMMIT,
 
     /* NML3Cfg hooks to the NMPlatform signals for link, addresses and routes.
@@ -168,8 +163,8 @@ typedef struct {
         struct {
             const NML3ConfigData *l3cd_old;
             const NML3ConfigData *l3cd_new;
-            bool                  commited;
-        } l3cd_changed;
+            bool                  l3cd_changed;
+        } commit;
 
         struct {
             NML3AcdAddrInfo info;
@@ -203,6 +198,7 @@ struct _NML3Cfg {
         const NMPObject          *plobj;
         const NMPObject          *plobj_next;
         int                       ifindex;
+        gboolean                  dns_route_added_x[2]; /* index with IS_IPv4 */
     } priv;
 
     /* NML3Cfg strongly cooperates with NMNetns. The latter is
